from scarv import scarv_queries

def getNonCodingPathogenic(exon_flank, ensembl_ftp, hgmd_vcf, annotation_table_fn):
    import pyranges as pr

    exons = scarv_queries.query_ensembl(ensembl_ftp, "exon")
    transcripts = scarv_queries.query_ensembl(ensembl_ftp, "transcript")
    introns = transcripts.subtract(exons)

    exons_extensions = exons.slack(exon_flank)\
                            .intersect(introns)

    hgmd_snvs = VCFtoPyRanges_HGMD(hgmd_vcf)

    hgmd_snvs_noncoding = hgmd_snvs.subtract(exons)\
                                   .subtract(exons_extensions)

    out = filterForRegulatory(hgmd_snvs_noncoding, annotation_table_fn)

    return out


# function that converts HGMD vcf to pybedtool object
# filters for: 1) snvs, 2) CLASS=="DM" or "DM?"
def VCFtoPyRanges_HGMD(vcf_file):
    from pysam import VariantFile

    HGMD_out_l = []
    bcf_in = VariantFile(vcf_file)

    for record in bcf_in.fetch():
        is_snv = (len(record.ref)==1 and len(record.alts[0])==1)
        is_dm = record.info['CLASS'] in ['DM', 'DM?']
        if (is_snv and is_dm):
            HGMD_out_l += [(record.contig, record.start, record.stop, record.alts[0])]

    HGMD_bed = scarv_queries.coordlist_to_pyranges(HGMD_out_l, entryNames = ["Chromosome", "Start", "End", "Alt"])
    HGMD_bed.Chromosome = ["chr" + str(chrom) for chrom in HGMD_bed.Chromosome]

    HGMD_bed = HGMD_bed[~HGMD_bed.as_df().duplicated()] # remove duplicates
    out = HGMD_bed.sort()
     
    return out


def filterForRegulatory(gr, annotation_table_fn):
    import pandas as pd
    import pyranges as pr

    annotation = pd.read_csv(annotation_table_fn, sep='\t')

    correct_type = (annotation['mutype'] == 'regulatory')
    regulatory_snvs_annotation = annotation[correct_type] 
    regulatory_snvs_annotation['chrom'] = ["chr" + str(chrom) for chrom in regulatory_snvs_annotation['chrom']]
    
    regulatory_snvs_annotation_unduplicated = regulatory_snvs_annotation[~regulatory_snvs_annotation[['chrom', 'pos','alt']].duplicated()]

    filtered_snvs = pd.merge(gr.as_df(), regulatory_snvs_annotation_unduplicated, left_on=['Chromosome', 'End', 'Alt'],\
        right_on=['chrom', 'pos','alt'], how='inner')

    return pr.PyRanges(filtered_snvs[['Chromosome', 'Start', 'End', 'Alt']])






def query_vcf(vcf_file, queries):
    from pysam import VariantFile

    data_dict = dict()
    for ix in range(len(queries)):
        data_dict[ix] = list()

    bcf_in = VariantFile(vcf_file)

    for record in bcf_in.fetch():
        for ix, query in enumerate(queries):
            to_add = process_record(record, query)
            if len(to_add) > 0:
                data_dict[ix].append(to_add)

    return(data_dict)



def process_record(record, query):
    AC_str = "AC_" + query['ancestry']
    AF_str = "AF_" + query['ancestry']
    AN_str = "AN_" + query['ancestry']

    # checks whether variant applies to ancestry
    if record.info[AC_str][0] == 0: 
        return ()

    if query['PASS'] is not None:
            if ("PASS" in list(record.filter)) != query['PASS']:
                return ()

    if query['variant_type'] is not None:
        if record.info['variant_type'] != query['variant_type']:
            return ()

    # flip ref & alt in MAF > 0.5
    if (record.info[AF_str][0] <= 0.5):
        out = (record.contig, record.start, record.stop, record.ref, record.alts[0],\
            str(record.info[AC_str][0]), str(record.info[AN_str]))
    else:
        out = (record.contig, record.start, record.stop, record.alts[0], record.ref,\
            str(record.info[AN_str] - record.info[AC_str][0]), str(record.info[AN_str]))

    return out



def get_reliable_sites(coverage_file, fail_vars, pop_split):
    import pyranges as pr

    covered_loci_list = filter_coverage(coverage_file)
    covered_loci = coordlist_to_pyranges(covered_loci_list)
    covered_loci_mgd = covered_loci.merge()

    reliable_sites = covered_loci_mgd.subtract(fail_vars)

    return reliable_sites



# assumes format chrX:pos, mean, median, over_1, over_5, over_10, over_15, over_20, over_25, over_30, over_50, over_100
def filter_coverage(coverage_file, pop_split):
    import gzip

    f = gzip.open(coverage_file, 'r') 
    headers = next(f).rstrip().split(b'\t')

    over_15_col = [ix for ix, col in enumerate(headers) if col==b"over_15"][0]
    over_100_col = [ix for ix, col in enumerate(headers) if col==b"over_100"][0]

    out_l = []
    for line in f:
        line_spl = line.decode().split('\t')
        chrom, pos = line_spl[0].split(':')

        over_15_threshold = 0.9
        if chrom == "chrX": 
            if (int(pos) >= 10001) or (int(pos) <= 2781479): # PAR1
                over_15_threshold = 0.9 
            elif (int(pos) >= 155701383) or (int(pos) <= 156030895): #PAR2
                over_15_threshold = 0.9
            else:
                over_15_threshold = (pop_split["XX"] * 0.9 + pop_split["XY"] * 0.5)/(pop_split["XX"] + pop_split["XY"])
 
        if (float(line_spl[over_15_col]) >= over_15_threshold) and (float(line_spl[over_100_col]) <= 0.1):
            pos = line_spl[0].split(':')
            out_l.append((pos[0], int(pos[1])-1, int(pos[1])))

    return out_l



def get_training_loci(reliable_sites, phyloP_bw, ensembl_ftp, nonsingleton_snvs):
    import pyranges as pr
    
    phylop_filtered_list = filter_phyloP(phyloP_bw)
    phylop_filtered = coordlist_to_pyranges(phylop_filtered_list)
    phylop_filtered_mgd = phylop_filtered.merge()

    exons = query_ensembl(ensembl_ftp, "exon")

    training_loci = reliable_sites.intersect(phylop_filtered_mgd)\
                                  .subtract(exons)\
                                  .subtract(nonsingleton_snvs)
    
    return training_loci



def filter_phyloP(phyloP_bw, phyloP_range=[-1.3,0]):
    import pyBigWig

    bw = pyBigWig.open(phyloP_bw)

    out_l = []
    for chrom in bw.chroms().keys():
        vals = bw.intervals(chrom)
        l = [(chrom, record[0], record[1]) for record in vals\
            if ((record[2] >= phyloP_range[0]) and (record[2] <= phyloP_range[1]))]
        out_l.extend(l)

    return out_l



# gtf format uses inclusive range 
def query_ensembl(ensembl_ftp, annot):
    import pandas as pd
    import pyranges as pr

    data = pd.read_csv(ensembl_ftp, sep='\t', header=None,\
        names=['Chromosome', 'source', 'type', 'Start', 'End', 'score', 'strand',\
        'phase', 'attributes'], skiprows=[i for i in range(5)], compression="gzip")
    data['Start'] = data['Start'] - 1 

    data_correct_type = data.loc[data['type']==annot]
    data_correct_type["Chromosome"] = ["chr" + str(chrom) for chrom in data_correct_type["Chromosome"]]

    data_correct_type_pr = pr.PyRanges(data_correct_type)

    return data_correct_type_pr



def query_sequence(gr, flank, genome, reference_fasta):
    import pandas as pd
    import numpy as np
    import pyranges as pr

    chr_list = gr.Chromosome[~gr.Chromosome.duplicated()]       # unique chromosomes, but in order

    ix = 0
    out = np.empty(shape=(np.sum(2 * gr.lengths() - 1), 4*flank+1, 5), dtype=np.int8)
    for chrom in chr_list:
        chrom_gr = pr.from_dict({'Chromosome': [chrom], 'Start': [0], 'End': [genome[chrom][1]]})
        chrom_seq = pr.get_fasta(chrom_gr, reference_fasta)[0]

        seqs = [chrom_seq[(pos.Start-flank):(pos.End+flank)] for pos in gr[chrom].as_df().itertuples()]
        seqs_upper = ["X".join(seq.upper()) for seq in seqs]
        seqs_upper_split = [[seq[i:(i+4*flank+1)] for i in np.arange(len(seq)-4*flank)] for seq in seqs_upper]
        seqs_upper_flattened = [item for sublist in seqs_upper_split for item in sublist]

        nucs = [list(seq) for seq in seqs_upper_flattened]
        nucs_df = pd.DataFrame(nucs)

        # one hot encode the sequences
        nucs_cat = nucs_df.apply(lambda x: pd.Categorical(x, categories = ['A', 'C', 'X', 'G', 'T']))
        nucs_bin = pd.get_dummies(nucs_cat)
        nucs_rshpd = np.array(nucs_bin, dtype=np.int8).reshape(nucs_bin.shape[0], 4*flank+1, 5) 

        n_seqs = nucs_rshpd.shape[0]
        out[ix:(ix+n_seqs)] = nucs_rshpd
        ix += n_seqs

    return out



def correct_refs(gr, snvs_pr, seq):
    import numpy as np
    import pandas as pd
    import pyranges as pr

    flank = seq.shape[1]//2

    gr_spl = gr.tile(1)
    row_ix = pd.Series(range(len(gr_spl)), name="id")
    gr_spl = gr_spl.insert(row_ix)

    gr_spl_on_snvs_hits = gr_spl.join(snvs_pr)
    
    anyHits = (gr_spl_on_snvs_hits.length != 0)
    
    if anyHits:
        gr_spl_hit_ids = gr_spl_on_snvs_hits.id

        nucs_df = pd.DataFrame(gr_spl_on_snvs_hits.ref)
        nucs_cat = nucs_df.apply(lambda x: pd.Categorical(x, categories = ['A', 'C', 'X', 'G', 'T']))
        
        refs_from_vcf_ohe = np.array(pd.get_dummies(nucs_cat))
        seq[gr_spl_hit_ids, flank] = refs_from_vcf_ohe
     
    return 
    

def split_data(training_regions, chrXnonPAR, singleton_variants, sequence, pop_split):
    import numpy as np
    import pandas as pd
    import pyranges as pr

    n_sites = training_regions.length
    flank = sequence.shape[1]//2

    training_sites = training_regions.tile(1)
    row_ix = pd.Series(range(n_sites), name="id")
    training_sites = training_sites.insert(row_ix)

    # address chrX and chrY
    training_sites = training_sites[training_sites.Chromosome != "chrY"]
    chrXnonPAR_id = training_sites.join(chrXnonPAR).id

    neutral_singleton_variants = singleton_variants.join(training_sites)

    AC = np.array(neutral_singleton_variants.ac, dtype=int)
    AC_skew = np.random.binomial(n=AC, p=0.6)                       # 60% of alternates dedicated to skewed splits
    AC_bal = AC - AC_skew                                           # remaining 40% to balanced split

    AN = np.repeat(2 * (pop_split['XY'] + pop_split['XX']), n_sites)    # In abuse of notation, AN refers to reference allele count 
    AN[chrXnonPAR_id] = pop_split['XY'] + 2 * pop_split['XX']

    AN[neutral_singleton_variants.id] = neutral_singleton_variants.an - 1   # -1 due to singleton

    AN_skew = np.random.binomial(AN, p=0.6)                         # 60% of references dedicated to skewed splits
    AN_rem = AN - AN_skew
    AN_bal = balance_data(AN_rem, AC_bal)                           # subset of remaining 40% for balanced split

    N_bal = np.concatenate([AN_bal, AC_bal])
    N_skew = np.concatenate([AN_skew, AC_skew])

    N_cal = np.random.binomial(N_skew, p=2/3)                       # 2:1 split in calibration set versus test set
    N_test = N_skew - N_cal

    alts_ohe = np.array(pd.get_dummies(pd.Categorical(neutral_singleton_variants.alt, categories = ['A', 'C', 'X', 'G', 'T'])))
    refs_ohe = sequence[:,flank,:]
    nuc = np.concatenate([refs_ohe, alts_ohe])

    indices = np.concatenate((np.arange(n_sites), neutral_singleton_variants.id), axis=0)

    return indices, nuc, np.c_[N_bal, N_cal, N_test]



def balance_data(AN, AC):
    import collections
    import numpy as np

    samples = np.random.choice(len(AN), sum(AC), p=AN/sum(AN))
    counts = collections.Counter(samples)
    
    keys = [int(key) for key in counts.keys()]
    vals = [int(val) for val in counts.values()]

    AN_out = np.repeat(0, len(AN))
    AN_out[keys] = vals

    return AN_out



def coordlist_to_pyranges (pos_l, entryNames = ["Chromosome", "Start", "End"]):
    assert len(pos_l[0]) == len(entryNames), "not enough entry names provided"

    import pyranges as pr 

    values_by_entry = zip(*pos_l)
    pos_dict = {entry: values for entry, values in zip(entryNames, values_by_entry)}
    pos_pr = pr.from_dict(pos_dict)

    return pos_pr 



def sample_loci_from_pr (gr, n_loci):
    import pyranges as pr
    import numpy as np

    N = gr.length
    gr_spl = gr.tile(1)

    ix = np.random.choice(range(N), n_loci, replace=False)
    sample_pr = pr.PyRanges(gr_spl.as_df().iloc[ix,:])

    return sample_pr.sort()



def load_variants(fn):
    import pyranges as pr

    gr = pr.read_bed(fn) 
    gr.columns = ["Chromosome", "Start", "End", "ref", "alt", "ac", "an"]
    gr.ac = gr.ac.astype(int)

    return gr



def writeToBed(gr, fn):
    import pandas as pd
    import os

    fn_tmp = fn + ".tmp"
    gr.as_df().to_csv(fn_tmp, header=False, index=False, sep='\t')
    
    os.system("sort-bed " + fn_tmp + " > " + fn)
    os.system("rm " + fn_tmp)

    return




from scarv import scarv_queries


def get_entropy (p):
    import numpy as np
    return np.nansum(-p * np.log2(p), axis=p.ndim-1)



def calibrate(uncalibrated_predictions, calibration_model, sequence):
    import numpy as np

    epsilon = 1e-9
    flank = sequence.shape[1]//2
    class_scores = np.log(uncalibrated_predictions)

    ref_score = np.nansum(class_scores * sequence[:, flank, :], axis=1)
    ref_score[~np.isfinite(ref_score)] = np.log(epsilon)                    # calibration model doesn't accept -np.inf
    ref_score_calibrated = calibration_model.transform(ref_score)
    ref_prob = np.apply_along_axis(lambda x: x * ref_score_calibrated, 0, sequence[:, flank, :])

    alternate_probs_unnormalised = np.exp(class_scores) * (1 - sequence[:, flank, :])
    normalisation_factor = (1 - ref_score_calibrated) / np.sum(alternate_probs_unnormalised, axis=1)
    normalisation_factor[np.isnan(normalisation_factor)] = 1                # division by zero occurs when all weight is on reference
    alternate_probs = np.apply_along_axis(lambda x: x * normalisation_factor, 0, alternate_probs_unnormalised)

    prediction = ref_prob + alternate_probs
    return prediction



def match_scores_to_ranges(gr, gr_reliable_mgd, obs_reliable_flat, exp_reliable_flat):
    import pyranges as pr
    import pandas as pd
    import numpy as np
    import collections

    gr_reliable_spl = gr_reliable_mgd.tile(1)

    o = pd.Series(data = obs_reliable_flat, name="observed")
    e = pd.Series(data = exp_reliable_flat, name="expected")

    gr_reliable_spl = gr_reliable_spl.insert(o)
    gr_reliable_spl = gr_reliable_spl.insert(e)

    index = pd.Series(range(len(gr)), name="id")
    gr = gr.insert(index) 

    hits = gr.join(gr_reliable_spl, suffix="_spl")
    out = hits.as_df().groupby(['id']).agg({'observed': "sum", 'expected': "sum"})

    countsByIndex = collections.Counter(hits.id)
    out['coverage'] = [countsByIndex[ix] if ix in countsByIndex.keys() else 0 for ix in out.index]    

    return out



def process_variants (variants, allele_counts, insertion=False):
    import pandas as pd
    import numpy as np

    df = variants.as_df()
    df = df.set_index('Start')

    df.alt = pd.Categorical(df.alt, categories=['A','C','X','G','T'])
    df.ref = pd.Categorical(df.ref, categories=['A','C','X','G','T'])

    alternate_alleles = pd.get_dummies(df.alt).mul(df.ac, axis=0)\
                                                   .groupby('Start')\
                                                   .sum()\
                                                   .to_numpy()

    reference_alleles = pd.get_dummies(df.ref).mul(df.an, axis=0)\
                                                   .groupby('Start')\
                                                   .first()\
                                                   .to_numpy()

    reference_alleles -= np.sum(alternate_alleles, axis=1)[:, np.newaxis]
    reference_alleles[reference_alleles < 0] = 0

    increment = 1 if insertion else 0
    allele_counts[2 * np.unique(df.index) + increment] = reference_alleles + alternate_alleles

    return allele_counts



def toPercentile(scores, percentiles):
    import pandas as pd
    import numpy as np

    # extend boundaries in case min/max extends beyond sampled min/max
    percentiles[0] = -np.inf
    percentiles[-1] = np.inf 

    assert len(np.unique(percentiles)) == len(percentiles), "Percentile values are non-unique"

    percentiles = pd.cut(scores, bins=percentiles, labels=np.arange(0.1,100.1,0.1), include_lowest=True)
    return percentiles



def percScoreToCumulativePercCountPlot(percScore, plt=None, x_upper=None, dashed=False):
    cumulativeCounts = getCumulativeCounts(percScore)
    plt = plotCumulativePercCountPlot(cumulativeCounts, plt, x_upper, dashed)

    return plt



def getCumulativeCounts(percentile_values):
    import collections
    import numpy as np

    percentileTally = collections.Counter([int(10*x) for x in percentile_values])
    addMissingPercentiles(percentileTally)

    counts = [count for perc, count in sorted(percentileTally.items())]
    countsCumulative = np.sum(counts) - np.cumsum(counts)

    return countsCumulative



def plotCumulativePercCountPlot(countsCumulative, plt=None, x_lower=None, dashed=False):
    import numpy as np

    percentiles = np.arange(0, 100.1, 0.1)

    if dashed:
        plt.plot(percentiles, countsCumulative, '--', linewidth=3)
    else:
        plt.plot(percentiles, countsCumulative, linewidth=3)

    plt.set_xlabel("Percentile", fontsize=23)
    plt.set_ylabel("Cumulative count", fontsize=23)

    plt.set_ylim(0, 1.1 * countsCumulative[0])

    if x_lower is not None:
        plt.set_xlim(x_lower, 100.1)
        plt.set_ylim(0, 1.1 * max([line.get_ydata()[x_lower * 10 - 1] for line in plt.lines]))

    return plt



def addMissingPercentiles(percentileCountDictionary):
    import numpy as np

    percentileSetComplete = set(np.arange(0, 1001, 1))
    percentilesPresent = set(percentileCountDictionary.keys())
    
    toAdd = percentileSetComplete - percentilesPresent
    for i in list(toAdd):
        percentileCountDictionary[i] = 0

    return None





from scarv import scarv_diagnostics

def prep_data(indices, y, multiplier, seq_in, expand=False):
    import numpy as np

    seq = seq_in[indices]
    if expand:
        ix = np.repeat(range(len(indices)), multiplier)
        seq_out = seq[ix]
        y_out = y[ix]
        return seq_out, y_out 
    else:
        ix = np.where(multiplier > 0)
        weight_out = multiplier[ix]
        seq_out = seq[ix]
        y_out = y[ix]
        return seq_out, y_out, weight_out



def fit_cnn(X, y):
    import keras
    import numpy as np

    flank = X.shape[1]//2

    learning_rate = 0.0001
    mini_batch_size = 64
    n_epochs = 1000
    patience = 5
    interval = 1

    model = create_cnn(flank=flank)

    # 90-10 split into training and validation
    ix = np.random.choice(range(X.shape[0]), round(0.9*X.shape[0]), replace=False)

    X_val, y_val = np.delete(X, ix, axis=0), np.delete(y, ix, axis=0)
    X_train, y_train = X[ix], y[ix]

    optim = keras.optimizers.Adam(lr = learning_rate)
    model.compile(optimizer=optim,loss='categorical_crossentropy', metrics=['accuracy'])

    ival = scarv_diagnostics.IntervalEvaluation(validation_data=(X_val, y_val), flank=flank, patience=patience, interval=interval)
    model.fit(x = X_train, y = y_train, epochs=n_epochs, batch_size = mini_batch_size, callbacks = [ival])

    return model 


def create_cnn(flank):
    import keras
    import keras_genomics

    model = keras.models.Sequential()

    model.add(keras_genomics.layers.RevCompConv1D(filters=100, kernel_size=5, input_shape=(2*flank+1, 5), 
        padding="same", activation="relu", kernel_initializer="he_normal"))

    model.add(keras_genomics.layers.RevCompConv1D(filters=100, kernel_size=5, padding="same",
        activation="relu", kernel_initializer="he_normal"))

    model.add(keras_genomics.layers.RevCompConv1D(filters=100, kernel_size=model.layers[-1].output_shape[1],
        activation="relu", kernel_initializer="he_normal"))

    model.add(keras_genomics.layers.RevCompConv1D(filters=3, kernel_size=model.layers[-1].output_shape[1],
                activation="softmax", kernel_initializer="he_normal"))
    
    model.add(keras.layers.Flatten())

    model.add(keras.layers.Lambda(sum_middle_elements))

    return model



def sum_middle_elements(x):
    from keras import backend as K

    A_and_C = x[:, 0:2,]
    X = K.sum(x[:, 2:4], axis=1, keepdims=True)
    G_and_T = x[:, 4:]
    return K.concatenate([A_and_C, X, G_and_T], axis=1)



def fit_calibration(model, seq, y, weights):
    # isotonic regression
    from sklearn.isotonic import IsotonicRegression 
    import numpy as np

    Z = np.log(model.predict(seq))

    flank = seq.shape[1]//2
    Z_ref = np.nansum(Z * seq[:, flank, :], axis=1).astype(float)
    Z_ref[np.isinf(Z_ref)] = -100 # isotonic regression does not allow -inf

    y_bin = 1 * (np.argmax(y, axis=1) == np.argmax(seq[:, flank, :], axis=1))

    ir = IsotonicRegression(y_min=0, y_max=1, out_of_bounds='clip')
    ir.fit(Z_ref, y_bin, sample_weight=weights)

    return ir








# scarv_classifier

def query_matching_variants(patho_SNVs, benign_SNVs, ratio, genome_segmentation, splice_sites):
    from scarv import scarv_queries
    import pyranges as pr
    import pandas as pd
    import numpy as np
    import collections
    import os

    patho_SNVs.Annot = "Intergenic"
    benign_SNVs.Annot = "Intergenic"

    for feature in genome_segmentation.keys():
        hits_benign = benign_SNVs.join(genome_segmentation[feature], how='left')
        benign_SNVs.Annot = [x if y=='-1' else y for x, y, in zip(hits_benign.Annot, hits_benign.Name)]

        hits_patho = patho_SNVs.join(genome_segmentation[feature], how='left')
        patho_SNVs.Annot = [x if y=='-1' else y for x, y, in zip(hits_patho.Annot, hits_patho.Name)]

    # annotate with distance to splice
    benign_SNVs = benign_SNVs.k_nearest(splice_sites, ties='first')[['Chromosome', 'Start', 'End', 'Annot', 'Distance']]
    benign_SNVs.Distance = benign_SNVs.Distance.abs()

    patho_SNVs = patho_SNVs.k_nearest(splice_sites, ties='first')[['Chromosome', 'Start', 'End', 'Annot', 'Distance']]
    patho_SNVs.Distance = patho_SNVs.Distance.abs()

    matched_benign_SNVs = pr.PyRanges()
    for SNV in patho_SNVs.as_df().iterrows():
        element = SNV[1].Annot
        distToSplice = SNV[1].Distance
        matched_benign_SNVs = sample_variants(element, distToSplice, matched_benign_SNVs, benign_SNVs, ratio)

    # remove duplicates and ensure distance of > 500 bp between benign variants
    matched_benign_SNVs_no_dups = matched_benign_SNVs.merge().tile(1)
    matched_benign_SNVs_no_dups_extd = matched_benign_SNVs_no_dups.slack(250).merge()

    while(len(matched_benign_SNVs_no_dups_extd) != len(matched_benign_SNVs_no_dups)):
        first_overlap = np.where(matched_benign_SNVs_no_dups_extd.lengths() > 501)[0][0]
        matched_benign_SNVs_no_dups = pr.PyRanges(matched_benign_SNVs_no_dups.as_df().drop(first_overlap))
        matched_benign_SNVs_no_dups_extd = matched_benign_SNVs_no_dups.slack(250).merge()

    return matched_benign_SNVs_no_dups



def sample_variants(element, distToSplice, sampled_variants, control_variants, ratio):
    import pyranges as pr
    import pandas as pd
    import numpy as np  

    correct_annotation = control_variants[control_variants.Annot==element].as_df()
    correct_annotation['distDiff'] = np.abs([x - distToSplice for x in correct_annotation.Distance])
    correct_annotation_std_by_distDiff = correct_annotation.sort_values('distDiff')

    new_samples = pr.PyRanges(correct_annotation_std_by_distDiff.iloc[:ratio])[['Chromosome', 'Start', 'End']]
    sampled_variants = pr.PyRanges(pd.concat([sampled_variants.as_df(), new_samples.as_df()]))

    return sampled_variants



def load_data (gr, fn, score_annotation_fns = {}, gene_annotation_fns = {}, pLI_association_annotation_fns = {}, superenhancer_annotation_fns = {}):
    import pyranges as pr
    import pandas as pd
    import numpy as np  

    for feature in score_annotation_fns.keys():
        print("Annotating with " + feature)
        gr = query_annotation(gr, fn, score_annotation_fns[feature], feature)
        print("Done")

    for feature in gene_annotation_fns.keys():
        print("Annotating with " + feature)
        gr = query_nearest_gene_feature(gr, gene_annotation_fns[feature], feature)
        print("Done")

    for feature in pLI_association_annotation_fns.keys():
        print("Annotating with " + feature)
        gr = annotate_pLI_associations(gr, pLI_association_annotation_fns[feature], [feature, feature + "_ind"])
        print("Done")

    for feature in superenhancer_annotation_fns.keys():
        print("Annotating with " + feature)
        gr = annotate_superenhancer(gr, superenhancer_annotation_fns[feature], feature)
        print("Done")

    return gr



def query_annotation (gr, gr_fn, annot_fn, name):
    import os
    import pyranges as pr
    import numpy as np

    query = os.popen("tabix " + annot_fn + " -R " + gr_fn).read().split()

    x = np.where(["chr" in x for x in query])[0]
    ncols = x[1] - x[0]

    gr_annot = pr.from_dict({'Chromosome': query[::ncols], 'Start': list(map(int, query[1::ncols])), 'End': list(map(int, query[2::ncols])), name: list(map(float, query[3::ncols]))})
    
    df_annot = gr_annot.as_df()
    df_annot['Chromosome'] = df_annot['Chromosome'].astype("object")

    gr_annot_max = pr.PyRanges(df_annot.groupby(['Chromosome','Start','End'])[name].max().dropna().reset_index())
    gr = annotate(gr, gr_annot_max, name)

    return gr



def annotate (gr, gr_annot, name):
    import numpy as np
    import pyranges as pr

    cols = list(gr.columns)
    new_cols = cols + [name]

    hits = gr.join(gr_annot, how='left').as_df()
    hits_no_dups = hits[~hits[['Chromosome', 'Start', 'End']].duplicated()].copy()
    hits_no_dups.loc[hits_no_dups.End_b == -1, name] = np.nan

    gr = pr.PyRanges(hits_no_dups)[new_cols]

    return gr



def query_nearest_gene_feature(gr, annot_fn, name):
    import pyranges as pr

    annot = pr.read_bed(annot_fn)
    annot.columns = ['Chromosome', 'Start', 'End', name]

    gr_annot = gr.k_nearest(annot, ties='first')[[name]]
    gr = annotate(gr, gr_annot, name)

    return gr



def annotate_pLI_associations (gr, pLI_annotation_fn, names):
    import numpy as np
    import pyranges as pr
    import pandas as pd

    cols = list(gr.columns)
    new_cols = cols + names

    pLI_annotation = pr.read_bed(pLI_annotation_fn)
    pLI_annotation.columns = ['Chromosome', 'Start', 'End', names[0]]

    hits = gr.join(pLI_annotation, how='left').as_df()
    hits.loc[hits['End_b'] == -1, names[0]] = np.nan
    hits[names[1]] = 1 - np.isnan(hits[names[0]]).astype(int)
    hits_no_dups = hits[~hits[['Chromosome', 'Start', 'End']].duplicated()].copy()

    gr = pr.PyRanges(hits_no_dups)[new_cols]

    return gr



def annotate_superenhancer (gr, superenhancer_annotation_fn, name):
    import numpy as np
    import pyranges as pr

    cols = list(gr.columns)
    new_cols = cols + [name]

    superenhancer_annotation = pr.read_bed(superenhancer_annotation_fn)
    superenhancer_annotation.columns = ['Chromosome', 'Start', 'End', name]

    hits = gr.join(superenhancer_annotation, how='left').as_df()
    hits_no_dups = hits[~hits[['Chromosome', 'Start', 'End']].duplicated()].copy()
    hits_no_dups.loc[hits_no_dups.End_b == -1, name] = 0

    gr = pr.PyRanges(hits_no_dups)[new_cols]

    return gr



def get_AUC(train_data, test_data):
    from xgboost import XGBClassifier
    from sklearn.metrics import roc_auc_score

    clf = XGBClassifier(eval_metric='logloss', use_label_encoder=False)
    clf.fit(train_data['X'], train_data['y'])

    preds_prob = clf.predict_proba(test_data['X'])[:,1]
    auc = roc_auc_score(test_data['y'], preds_prob)

    return auc



def cross_validation_k_fold(X, Y, k):
    from xgboost import XGBClassifier
    import numpy as np

    folds = np.random.choice(np.arange(k), size=X.shape[0], replace=True)
    preds = np.empty(X.shape[0])

    for i in range(k):
        X_train, X_test = X[folds!=i], X[folds==i]
        Y_train, Y_test = Y[folds!=i], Y[folds==i]

        clf = XGBClassifier(eval_metric='logloss', use_label_encoder=False)
        clf.fit(X_train, Y_train)

        preds[folds==i] = clf.predict_proba(X_test)[:,1]

    return preds




def split(x, f):
    import itertools

    count = max(f) + 1
    return tuple( list(itertools.compress(x, (el == i for el in f))) for i in range(count) )  


def reliability_plot (confidence, accuracy):
    import matplotlib.pyplot as plt

    min_acc = min(accuracy)
    min_conf = min(confidence)

    plt.plot(confidence, accuracy, color="blue")
    plt.plot([0,1],[0,1])
    
    plt.ylim((min(min_acc, min_conf), 1))
    plt.xlim((min(min_acc, min_conf), 1))
    
    plt.xlabel('Confidence')
    plt.ylabel('Accuracy')

    print(plt.show())

    return 


def expected_calibration_error(n_bins, preds, y_mat, weights, plot=True):
    import collections
    import numpy as np
    import pandas as pd

    y_hat = np.argmax(preds, axis=1)
    p_hat = np.max(preds, axis=1)

    binned_preds = pd.qcut(p_hat, n_bins, labels=False, duplicates='drop').astype(int)
    bin_sizes = collections.Counter(binned_preds)

    y = np.argmax(y_mat, axis=1)
    n = len(y_hat)

    l1 = list(map(sum, split((y==y_hat) * weights, binned_preds))) 
    l2 = list(map(sum, split(weights, binned_preds))) 
    acc_by_bin = [x/y for x, y in zip(l1, l2) if y > 0]

    l3 = list(map(sum, split(p_hat * weights, binned_preds)))
    conf_by_bin = [x/y for x, y in zip(l3, l2) if y > 0]

    if plot:
        reliability_plot(conf_by_bin, acc_by_bin)

    ECE = np.sum([abs(x-y)*z for x,y,z in zip(acc_by_bin, conf_by_bin, l2)])/sum(l2)

    return ECE


def hand_till_auc(preds, y_test, weights=None):
    import collections
    import numpy as np

    if weights is None:
        weights = np.repeat(1, preds.shape[0])

    running_total_AUC = 0

    pairs = [[i, j] for i in range(4) for j  in range(4) if i!=j]
    for i, j in pairs:
        # select indices for which either i or j is observed
        ix = (y_test[:,i] == 1) | (y_test[:,j] == 1)

        # determine the prediction for class i and whether i occurred 
        preds_i = preds[ix, i]
        y = y_test[ix, i]
        w = weights[ix]

        # determine the order of the predictions
        order = np.argsort(preds_i)
        preds_i_std = preds_i[order]
        y_std = y[order]
        w_std = w[order]

        start_ix = (np.concatenate(([0], np.cumsum(w_std))) + 1)[:-1]
        end_ix = start_ix + w_std

        S_i = 0
        for i in range(len(start_ix)):
            if y_std[i] == 1:
                S_i = S_i + sum(range(start_ix[i], end_ix[i]))

        n_i = sum(w_std[y_std == 1])
        n_j = sum(w_std[y_std == 0])

        # calculate A(i|j) is order to prevent overflow
        A_ij = (S_i/n_i)/n_j - (n_i + 1)/(2 * n_j)
        running_total_AUC += A_ij

    AUC = running_total_AUC / len(pairs)
    return AUC



def max_auc(preds, weights=None):
    import numpy as np

    if weights is None:
        weights = np.repeat(1, preds.shape[0])

    output = np.empty(shape=(preds.shape[0], 5), dtype=int)

    for i in range(preds.shape[0]):
        output[i,:] = np.random.multinomial(weights[i], preds[i,:], size=1)
    
    output_extd = output[np.repeat(np.arange(output.shape[0]), np.sum(output > 0, axis=1))] 
    weight = np.sum(output_extd, axis=1)
    output_binary = (output_extd > 0)*1
    p = preds[np.repeat(np.arange(output.shape[0]), np.sum(output > 0, axis=1))]

    AUC = hand_till_auc(p, output_binary, weight)
    return AUC



from keras.callbacks import Callback
class IntervalEvaluation(Callback):

    def __init__(self, validation_data, flank, interval=1, wait=0, patience=1):
        super(Callback, self).__init__()

        self.interval = interval
        self.patience = patience
        self.wait = wait
        self.flank = flank
        self.X_val, self.y_val = validation_data
        self.auc = 0
        self.best_weights = None

    # update auc if maximum
    # if not, wait incremented
    # if wait equals patience, training is stopped
    def on_epoch_end(self, epoch, logs={}):
        if epoch % self.interval == 0:
            preds = self.model.predict(self.X_val)
            score = hand_till_auc(preds, self.y_val)
            if score <= self.auc:
                if self.wait >= self.patience:
                    self.model.stop_training = True
                    print('Restoring model weights from the end of the best epoch.')
                    self.model.set_weights(self.best_weights)
                else:
                    self.wait += 1
            else:
                self.best_weights = self.model.get_weights()
                self.auc = score
                self.wait = 0
            print(score)







# functions to correct multiallelic inverted sites (inconsistencies exist in allele counts and refs)
def insert_corrected_snvs (correction, snvs):
    import pyranges as pr
    import pandas as pd

    no_correction_required = snvs.subtract(correction)
    no_correction_required.columns = no_correction_required.columns.str.replace("old_ref", "ref")

    inserted_df = pd.concat([no_correction_required.as_df(), correction.as_df()], axis=0)
    sorted_gr = pr.PyRanges(inserted_df).sort()

    sorted_gr = sorted_gr[sorted_gr.ac.astype(int)>0]
    
    return sorted_gr



def correct_inverted_multiallelics(snvs):
    snvs.columns = snvs.columns.str.replace("ref", "old_ref")

    multiallelic_loci = get_multiallelic_sites (snvs)
    inverted_multiallelic_sites = filter_for_inverted_sites (multiallelic_loci, snvs)
    annot_with_major_allele = get_major_alleles(inverted_multiallelic_sites, snvs)
    out = correct_allele_count (annot_with_major_allele)

    out = out.drop(['ix', 'old_ref'])

    return out



# find multiallelic sites
def get_multiallelic_sites (snvs):
    loci = snvs[['Chromosome', 'Start', 'End']].as_df()

    multiallelic_loci = loci[loci.duplicated(keep='first')].drop_duplicates()\
                                                           .reset_index(drop=True)
    return multiallelic_loci



# find inverted multiallelic sites
def filter_for_inverted_sites (multiallelic_loci, snvs):
    import pyranges as pr
    import pandas as pd

    multiallelic_loci_gr = pr.PyRanges(multiallelic_loci)

    index = pd.Series(range(len(multiallelic_loci_gr)), name='ix')
    multiallelic_loci_gr = multiallelic_loci_gr.insert(index)

    multiallelic_loci_annot = multiallelic_loci_gr.join(snvs)
    is_inverted_indicator = multiallelic_loci_annot.as_df().groupby('ix')['old_ref'].agg(is_inverted)

    inverted_multiallelic_sites = pr.PyRanges(multiallelic_loci.loc[is_inverted_indicator])

    return inverted_multiallelic_sites



def is_inverted (alleles):
    import numpy as np

    return len(np.unique(alleles)) != 1



# query correct refs
def get_major_alleles (inverted_multiallelic_sites, snvs):
    import pandas as pd

    index_inverted = pd.Series(range(len(inverted_multiallelic_sites)), name='ix')
    inverted_multiallelic_sites = inverted_multiallelic_sites.insert(index_inverted)

    out = inverted_multiallelic_sites.join(snvs).drop(['Start_b', 'End_b'])
    correct_refs = out.as_df().groupby('ix').apply(get_correct_ref)
    correct_refs_expanded = pd.Series(correct_refs[out.ix], name="ref")

    out = out.insert(correct_refs_expanded)

    return out



def get_correct_ref (x):
    import pandas as pd
    import numpy as np

    refs = np.array(x['old_ref'])
    alts = np.array(x['alt'])
    
    hg38_ref = np.array(list(set(refs).intersection(set(alts))))
    inversion_ix = np.where([alt in hg38_ref for alt in alts])[0]
    correct_ref = refs[inversion_ix][0]
    
    return correct_ref



# correct AC
def correct_allele_count (out):
    import numpy as np
    import pyranges as pr

    ac_correction = out[out.old_ref!=out.ref].as_df().groupby('ix')['ac'].sum().reset_index(drop=True)

    tmpDf = out.as_df()
    tmpDf = tmpDf.astype({"ac": int})

    tmpDf.loc[tmpDf.old_ref==tmpDf.ref, 'ac'] -= np.array(ac_correction.astype(int))

    out = pr.PyRanges(tmpDf)

    return out
